/*********************************************************************
DESCRIPTION :   The TTCdecoder receives a subset of the 
outputs of the TTCrx chip and generate the TTC commands.

/* On Mon, 3 May 2004, Jeremiah Mans wrote:

> We should _ignore_ the bits BRCST<2> and BRCST<4>.  These bits cause
> specific behavior on the FrontEnd.  We may or may not want to associate
> this behavior with any of these messages, so we should conside these
> bits as "don't care".  We are then left with
>
> 765X3XXX
>
> since the low two bits are also special -- EC reset and BC reset.
>
> I suggest this mapping:
>                          765X3XXX
> BCZero                   000X1XXX
> ResetOrbitCounter        001X1XXX
> ReSync                   010X1XXX
> HardReset                011X1XXX
> Start                    100X1XXX
> Stop                     101X1XXX
> Test Enable              001X0XXX
> Private Gap              010X0XXX
> Private Orbit            011X0XXX
> ResetEventCounter        XXXXXX1X
> No Operation             000X0XXX
>
> This mapping leaves six codes available for any last-minute functionality.


It also generate locally the BCN, EvN, OrN.
*/  /*
NB: all TTC_input are synchronous with TTC_CK40Des1 and at some point
the signals must cross to the domain of the system clock (clk).
The best strategy seems to have only 1-bit 1-tick signals (not busses)
crossing the clock domains; it would be a problem if a bus
crosses a clock domain while not all its lines switch at the
same time.  This implies that:
- the decoding of TTC_Brcst is confined in the domain of TTC_CK40Des1
- the pulses crossing the domains are:
      TTC_L1A, TTC_BCntRes,  TTC_EvCntRes, TTC_Hard_rst,
      TTC_Soft_rst, TTC_Start, TTC_Stop, TTC_BC0, TTC_StRq;
- BCN, EvN, OrN will be generated in the domain of clk.
- LastTTCcommand is a quasi-static signal, no special care.
Note that TTC_CK40Des1 has a jitter of ~ 1 ns pk-pk,
this could create an L1A pulse shorter than a cycle of clk;
we do a digital edge-detection with the 80MHz clk2x.
*****************************************************************************/

`include "define.v"
module TTCdecoder
(   
 input clk, clk2x, TTC_CK40Des1, reset, VME_ResCTrEvN,
 input [7:2] TTC_Brcst,
 input tTTC_BCntRes,  tTTC_EvCntRes, TTC_L1Accept, TTC_BrcstStr,   
 input [`BCN_WIDTH-1:0] MaxBCN, SetBCN,  
 output reg [15:0]
        TTC_test0, TTC_test1, TTC_test2, TTC_test3,
        TTC_test4, TTC_test5, TTC_test6, TTC_test7,
        TTC_test8, TTC_test9, TTC_testA, TTC_testB,
        TTC_testC, TTC_testD, TTC_testE, TTC_testF,
 output TTC_L1A,  TTC_Hard_rst, TTC_Soft_rst, TTC_Resync,
        TTC_TestEnable, TTC_Start, TTC_Stop, TTC_BC0,
        TTC_StRq, TTC_AsyncFifoRst, TTC_CalibTrig,
 output reg [7:0] LastTTCcommand = 'h00, 
 output reg [`EVN_WIDTH-1:0] EvN, CTrEvN,  // CalibTrigger Event Number
 output reg [`BCN_WIDTH-1:0] BCN,
 output reg [`OrN_WIDTH-1:0] OrN, 
 output reg BunchEr        );

 parameter 
  TTC0                     = 4'h0,
  TTC_BC0_code             = 4'h1,
  TTC_TestEnable_code      = 4'h2,
  TTC_RstOrbitCntr_code    = 4'h3,
  TTC_PrivateGap_code      = 4'h4,
  TTC_Resync_code          = 4'h5,
  TTC_PrivateOrbit_code    = 4'h6,
  TTC_Hard_rst_code        = 4'h7,
  TTC_Soft_rst_code        = 4'h8,   
  TTC_Start_code           = 4'h9,
  TTC_StRq_code            = 4'hA,  
  TTC_Stop_code            = 4'hB,
  TTC_AsyncFifoRst_code    = 4'hC,
  TTC_CalibTrig_code       = 4'hD,
  TTCE                     = 4'hE,
  TTCF                     = 4'hF;  
  
 wire [3:0] TTC_Broadcast_Bits = {TTC_Brcst[7],TTC_Brcst[6],TTC_Brcst[5],TTC_Brcst[3]};
 wire TTC_RstOrbitCntr; // this command is used only in this module
 // internal regs in the domain of TTC_CK40Des1 (t-prefix)
 reg  tRes, tTTC_Hard_rst, tTTC_Soft_rst, tTTC_Resync;
 reg  tTTC_TestEnable, tTTC_RstOrbitCntr, tTTC_CalibTrig;  
 reg  tTTC_Start, tTTC_Stop, tTTC_BC0, tTTC_StRq, tTTC_AsyncFifoRst;

 always @ (posedge TTC_CK40Des1) begin // test Brcst bus
    tRes <= reset; // synchronize reset to TTC_CK40Des1 
    if (tRes) begin 
      TTC_test0 <= 0;  TTC_test1 <= 0;  TTC_test2 <= 0;  TTC_test3 <= 0;
      TTC_test4 <= 0;  TTC_test5 <= 0;  TTC_test6 <= 0;  TTC_test7 <= 0;
      TTC_test8 <= 0;  TTC_test9 <= 0;  TTC_testA <= 0;  TTC_testB <= 0;
      TTC_testC <= 0;  TTC_testD <= 0;  TTC_testE <= 0;  TTC_testF <= 0;
    end
    else if (TTC_BrcstStr)   // Validate brcst bus
         case (TTC_Broadcast_Bits)
            0: TTC_test0 <= TTC_test0 + 1;  
            1: TTC_test1 <= TTC_test1 + 1;
            2: TTC_test2 <= TTC_test2 + 1;
            3: TTC_test3 <= TTC_test3 + 1;
            4: TTC_test4 <= TTC_test4 + 1;
            5: TTC_test5 <= TTC_test5 + 1;
            6: TTC_test6 <= TTC_test6 + 1;
            7: TTC_test7 <= TTC_test7 + 1;
            8: TTC_test8 <= TTC_test8 + 1;
            9: TTC_test9 <= TTC_test9 + 1;
           10: TTC_testA <= TTC_testA + 1;
           11: TTC_testB <= TTC_testB + 1;
           12: TTC_testC <= TTC_testC + 1;
           13: TTC_testD <= TTC_testD + 1;
           14: TTC_testE <= TTC_testE + 1;
           15: TTC_testF <= TTC_testF + 1;
         endcase
 end  // always
   
 always @ (posedge TTC_CK40Des1)   begin     
    if (TTC_BrcstStr)  begin // Validate brcst bus
       LastTTCcommand <= {TTC_Brcst, tTTC_EvCntRes, tTTC_BCntRes};
       case (TTC_Broadcast_Bits[3:0])
          TTC_Hard_rst_code      :  tTTC_Hard_rst      <=1;
          TTC_Soft_rst_code      :  tTTC_Soft_rst      <=1;
          TTC_Resync_code        :  tTTC_Resync        <=1;
          TTC_TestEnable_code    :  tTTC_TestEnable    <=1;
          TTC_RstOrbitCntr_code  :  tTTC_RstOrbitCntr  <=1;
          TTC_Stop_code          :  tTTC_Stop          <=1;     
          TTC_Start_code         :  tTTC_Start         <=1;
          TTC_BC0_code           :  tTTC_BC0           <=1; 
          TTC_StRq_code          :  tTTC_StRq          <=1; 
          TTC_AsyncFifoRst_code  :  tTTC_AsyncFifoRst  <=1; 
          TTC_CalibTrig_code     :  tTTC_CalibTrig     <=1;  
       endcase 
    end
    else begin    // not a broadcast cycle
       tTTC_Hard_rst<=0; tTTC_BC0<=0;  tTTC_Soft_rst<=0;   
       tTTC_Start<=0;    tTTC_Stop<=0; tTTC_RstOrbitCntr<=0; tTTC_CalibTrig<=0;  
       tTTC_Resync<=0;   tTTC_StRq<=0; tTTC_AsyncFifoRst<=0; tTTC_TestEnable<=0; 
    end  
 end   // always   

 TTCsynchronizer L1Async  (clk, clk2x, TTC_L1Accept,      TTC_L1A);
 TTCsynchronizer HRSTsync (clk, clk2x, tTTC_Hard_rst,     TTC_Hard_rst);
 TTCsynchronizer SRSTsync (clk, clk2x, tTTC_Soft_rst,     TTC_Soft_rst);
 TTCsynchronizer RSYNsync (clk, clk2x, tTTC_Resync,       TTC_Resync);
 TTCsynchronizer TSTEsync (clk, clk2x, tTTC_TestEnable,   TTC_TestEnable);
 TTCsynchronizer RORBsync (clk, clk2x, tTTC_RstOrbitCntr, TTC_RstOrbitCntr);
 TTCsynchronizer STRTsync (clk, clk2x, tTTC_Start,        TTC_Start);
 TTCsynchronizer STOPsync (clk, clk2x, tTTC_Stop,         TTC_Stop);
 TTCsynchronizer BC0sync  (clk, clk2x, tTTC_BC0,          TTC_BC0);
 TTCsynchronizer STRQsync (clk, clk2x, tTTC_StRq,         TTC_StRq);
 TTCsynchronizer BCRsync  (clk, clk2x, tTTC_BCntRes,      TTC_BCntRes);
 TTCsynchronizer ECRsync  (clk, clk2x, tTTC_EvCntRes,     TTC_EvCntRes);
 TTCsynchronizer AFRsync  (clk, clk2x, tTTC_AsyncFifoRst, TTC_AsyncFifoRst);
 TTCsynchronizer CTRsync  (clk, clk2x, tTTC_CalibTrig,    TTC_CalibTrig);

 always @ (posedge clk)   begin    // here BCN, EvN, OrN are generated      
    if (TTC_BCntRes) begin
         BCN     <= (SetBCN + 1);
         BunchEr <= (BCN != SetBCN);
    end
    else if (BCN != MaxBCN) BCN <= BCN +1;  
         else               BCN <= 0;
    if (TTC_EvCntRes)       EvN <= 1;   // the first L1A counted as 1
    else if (TTC_L1A)       EvN <= EvN + 1; // should match TTCrx EvCnt
    if (VME_ResCTrEvN)      CTrEvN <= 1;   // the first CalibTrig counted as 1
    else if (TTC_CalibTrig) CTrEvN <= CTrEvN + 1; 
    if (TTC_RstOrbitCntr)   OrN <=0;       
    else if (TTC_BC0)       OrN <= OrN +1; 
 end   // always       
endmodule  