`timescale 1ns / 1ps
module clk_change
  ( 
  input  clk2x,  clk,
  input  [17:0] Input_Data,
  output [10:0] QIE0data, QIE1data, QIE2data,
  output BZero, // = FE Beam_Zero; It won't be at BCN=0. We want to monitor it.
  output FE_DataFlag, FE_CntrlFlag,
  output reg reset_TP   );

  // word1_** is the higher portion of the transmitted word
  // word0_** is the lower  portion of the transmitted word
  wire [17:0]    latch80;
  reg  [17:0]    word1_l1_80,  word0_l1_80;
  reg  [17:0]    word1_l2_80;
  reg  [17:16]   word1_l3_80,  word0_l3_80; // only Er, DV used here.
  reg  [31:0]    GOL_InCopy;
  reg  clk2x_en;   // used instead of LSB for Er and DV
  wire LSB, flag1, flag2;
  reg  LSB_2d, LSB_d, LSBerror, LSBerror_D, LSBerror_40;
  reg  RxEr,  RxDV,   CapIDError,  TPres1, TPres2, TPres3, TPres4;
  reg  [1:0]  CapID_delayed;
  
  assign latch80 = Input_Data; // in case of Fmax problems, latch80 coud be registred 
  assign LSB     = latch80[0]; // NB: LSB toggles only when correct data are received
  
  always@(posedge clk2x) begin
    if (LSB == 1)   word1_l1_80[15:0] <= latch80[15:0];
    else            word0_l1_80[15:0] <= latch80[15:0]; 
    // here I monitor toggling of LSB
    LSB_d <= LSB; LSB_2d <= LSB_d;
    if (LSB_2d == LSB_d) LSBerror <= 1; 
    else                 LSBerror <= 0; 
    LSBerror_D <= LSBerror; 
    // when no correct data => no correct LSB, but we still want proper Er and DV:
    // generate locally clk2x_en to keep it on different lines from clk  
    clk2x_en <= ~clk2x_en; 
    if(clk2x_en)  word1_l1_80[17:16] <= latch80[17:16];
      else        word0_l1_80[17:16] <= latch80[17:16];
          
     word1_l2_80         <= word1_l1_80;  // extra delay to align the 2 half words.
     word1_l3_80[17:16]  <= word1_l2_80[17:16];  // delay to stratch Er=1 or DV=0 
     word0_l3_80[17:16]  <= word0_l1_80[17:16];    
  end  // always

// ************* CLOCK BOUNDARY : clk2x --> clk  **************

  always@(posedge clk) begin 
   // here we re-built the original input of the GOL chip:
   GOL_InCopy <= {word0_l1_80[15:0], word1_l2_80[15:0]};
   // Error if one or both words are error (stretched one extra cycle)
   RxEr <= word1_l3_80[17] || word0_l3_80[17] || word1_l2_80[17]  || word0_l1_80[17];
   // DataValid only if both words are valid (stretched one extra cycle)
   RxDV <= word1_l3_80[16] && word0_l3_80[16] && word1_l2_80[16]  && word0_l1_80[16]; 
   // Generation of CapID_Error:
   // we store a CapID pair for comparison and check that they all rotate togheter
   CapID_delayed <= GOL_InCopy[8:7]; 
   if( ((CapID_delayed + 2'b01) != GOL_InCopy[8:7])
       || (GOL_InCopy[8:7]!= GOL_InCopy[6:5])
       || (GOL_InCopy[8:7]!= GOL_InCopy[4:3])       )
             CapIDError <= 'd1;
       else  CapIDError <= 'd0;

    // stretch LSB error:
    LSBerror_40 <= LSBerror || LSBerror_D;
/*  long reset for the TriggerPath in case of the following:
    data non valid; CapID error ; orbit message;  */
    if (({flag1,flag2} != 2'b01)||(CapIDError)||(GOL_InCopy[2:1]!=2'b01))                                                     
               TPres1     <= 1;
        else   TPres1     <= 0;
    TPres4 <= TPres3; TPres3 <= TPres2; TPres2 <= TPres1;
    reset_TP <= TPres1 || TPres2 || TPres3 || TPres4;  //stratch
  end  // always                                    


    // combinatorial tables to generate flag1 and flag2
   FlagTables  FlagTables
    ( .b17(flag1), .b16(flag2), .Er(RxEr), .DV(RxDV), .LSB_er(LSBerror_40) );    

     // here we want to reformat and obtain:
     // QIExdata = Link_ER, Link_DV, CapID[1:0], Range[1:0], Mantissa[4:0]}
  assign QIE0data = {flag1, flag2, GOL_InCopy[8:7], GOL_InCopy[31:30], GOL_InCopy[29:25]}; 
  assign QIE1data = {flag1, flag2, GOL_InCopy[6:5], GOL_InCopy[23:22], GOL_InCopy[21:17]};
  assign QIE2data = {flag1, flag2, GOL_InCopy[4:3], GOL_InCopy[15:14], GOL_InCopy[13:9]};
       // NB: bit 10 and 9 of QIE*data are the same => possible warnings from synthesis.       
  assign  BZero        = GOL_InCopy[24];
  assign  FE_DataFlag  = GOL_InCopy[2];
  assign  FE_CntrlFlag = GOL_InCopy[1];

endmodule